#include "rrparser.h"
#include <iostream>

using namespace std;

///////////////////////////////////////////////////////////////////////////////////////////
// ==================================== class RRParser ====================================

RRParser::RRParser(CBuffer& buffer, Handler& handler)
: buffer_(buffer)
, error_(rrp_no_error)
, handler_(handler)
{
}

RRParser::~RRParser()
{
}

bool RRParser::Parse(RRData* pRRData)
{
	bool bRet;
	if (error_ != rrp_part_header && error_ != rrp_part_body){
		bRet = ParseString(pRRData);
		if( !bRet )
			return false;
	}

	if (error_ != rrp_part_body){
		bRet = ParseHeaders(pRRData);
		if( !bRet )
			return false;
	}

	Handler::t_handlerType type = handler_.getCommandType(pRRData->command.name);

	if (type == Handler::nocommand){
		error_ = rrp_request_error;
		return false;
	}

	if( type == Handler::withbody ) {
		bRet = ParseBody(pRRData);
		if( !bRet )
			return false;
	}

	return true;
}

bool RRParser::ParseHeaders(RRData* pRRData)
{
	string line;
	pRRData->headers.clear();

	if(!buffer_.is_online()){
		error_ = rrp_buffer_error;
		return false;
	}

	int k = 0;
	while(true) {
		line = "";
		buffer_.safegetline(line);

		k++;
		if (k > MAX_HEADERS_COUNT){
			error_ = rrp_request_error;
			return false;
		}

		if( buffer_.has_error() ) {
			if (buffer_.is_online()){
				if (buffer_.getsize() >= MAX_LINE_SIZE){
					buffer_.clear();
				}
				error_ = rrp_part_header;
				return false;
			} else {
				error_ = rrp_buffer_error;
				return false;
			}
		}

		if( line.size() >= MAX_LINE_SIZE ) {
			continue;
		}

		if( line.empty() )
			break;

		RRData::t_header header;
		if( !ParseHeader(line, header) ) {
			continue;
		}

		pRRData->headers.insert(header);
		//cout << "Parsed header: " << header.first.c_str() << ":" << header.second.c_str() << endl;
		if( pRRData->headers.size() > MAX_HEADERS_COUNT ) {
			error_ = rrp_request_error;
			return false;
		}
	}

	error_ = rrp_no_error;
	return true;
}

bool RRParser::ParseBody(RRData* pRRData)
{
	pRRData->body.clear();
	RRData::t_headers::const_iterator i;
	RRData::t_headers::const_iterator e = pRRData->headers.end();
	i = pRRData->headers.find("Content-Length");
	if( i == e ) {
		error_ = rrp_request_error;
		return false;
	}
	const int body_size = atoi(i->second.c_str());
	if (body_size == 0){
		error_ = rrp_no_error;
		return true;
	}

	char* buf = new char[body_size];

	const int size_read = buffer_.read(buf, body_size, body_size);
	pRRData->body.clear();
	//copy(buf, buf + size_read, back_inserter(pRRData->body));
	//cout << "here" << endl;
	pRRData->body.append(buf, size_read);
	delete[] buf;
	if( size_read != body_size ) {
		buffer_.rollback(1);
		error_ = rrp_part_body;
		return false;
	}
	//cout << "Parsed body: " << pRRData->body.c_str() << endl;
	error_ = rrp_no_error;
	return true;
}

bool RRParser::ParseProtocolVersion(const string& line, RRData* pRRData)
{
	string::const_iterator i = line.begin();
	string::const_iterator e = line.end();

	if( i == e || !isalpha((unsigned char)*i) ) 
		return false;
	string protocol = "";
	for(; i != e && isalpha((unsigned char)*i); protocol += *i++);
	if( i == e || *i != '/' )
		return false;
	++i;

	if( i == e || !isdigit((unsigned char)*i) )
		return false;
	int major(0);
	for(; i != e && isdigit((unsigned char)*i); major = 10 * major + *i++ - '0');
	if( i == e || *i != '.' )
		return false;
	++i;

	if( i == e || !isdigit((unsigned char)*i) ) 
		return false;
	int minor(0);
	for(; i != e && isdigit((unsigned char)*i); minor = 10 * minor + *i++ - '0');
	if( i != e )
		return false;

	pRRData->protocol.name = protocol;
	pRRData->protocol.version.major = major;
	pRRData->protocol.version.minor = minor;
	return true;
}

bool RRParser::ParseHeader(std::string line, RRData::t_header& header)
{
	string word1;
	string::const_iterator b;
	string::const_iterator i;
	string::const_iterator e;
	string::iterator j;
	word1 = SplitWord(line);
	b = word1.begin();
	i = b;
	e = word1.end();
	i = find(i, e, ':');
	if( i == e ) 
		return false;
	string key(b, i);
	++i;
	
	j = line.begin();
	j = find_if(j, line.end(), NotSpace);
	
	string value = string(i, e);
	if (j != e) value += string(j, line.end());
	header.first = key;
	header.second = value;
	return true;
}

vector< string > RRParser::Split(const string& line)
{
	vector< string > ret;

	string::const_iterator i, j, e;
	i = line.begin();
	e = line.end();

	while(i != e) {
		i = find_if(i, e, NotSpace);
		j = find_if(i, e, Space);
		if(i != j) {
			ret.push_back(string(i, j));
			i = j;
		}
	}

	return ret;
}

string RRParser::SplitWord(string& line)
{
	string ret;
	string::const_iterator i = line.begin();
	string::const_iterator e = line.end();
	string::const_iterator j;
	i = find_if(i, e, NotSpace);
	j = find_if(i, e, Space);
	if( i != j ) {
		ret = string(i, j);
		line = string(j, e);
	}
	return ret;
}

///////////////////////////////////////////////////////////////////////////////////////////
// ================================= class RRServerParser =================================

bool RRServerParser::ParseString(RRData* pRRData)
{
	string line;
	buffer_.safegetline(line);

	if(!buffer_.is_online()){
		error_ = rrp_buffer_error;
		return false;
	}
	
	if( buffer_.has_error() ) {
		if( buffer_.getsize() > MAX_LINE_SIZE ) {
			buffer_.getline(line);
			pRRData->command.name = line;
			error_ = rrp_request_error;
			return false;
		} else {
			error_ = rrp_part_command;
			return false;
		}
	} else if( line.size() > MAX_LINE_SIZE ) {
		pRRData->command.name = line;
		error_ = rrp_request_error;
		return false;
	} else {
		pRRData->command.name = line;
		vector< string > words = Split(line);
		if(words.size() < 2) {
			error_ = rrp_request_error;
			return false;
		}
		if(words.size() == 2) {
			if( !ParseProtocolVersion(words[1], pRRData) ) {
				error_ = rrp_request_error;
				return false;
			}
			pRRData->command.name = words[0];
			pRRData->command.parameter = "";
		} else {
			if( !ParseProtocolVersion(words[2], pRRData) ) {
				error_ = rrp_request_error;
				return false;
			}
			pRRData->command.name = words[0];
			pRRData->command.parameter = words[1];
		}
		//cout << "Recieved command: " << pRRData->command << ' ' << pRRData->protocol << endl;
		error_ = rrp_no_error;
		
		return true;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
// ================================= class RRClientParser =================================

bool RRClientParser::ParseString(RRData* pRRData)
{
	string line;
	buffer_.safegetline(line);

	if(!buffer_.is_online()){
		error_ = rrp_buffer_error;
		return false;
	}

	if( buffer_.has_error() ) {
		if( buffer_.getsize() > MAX_LINE_SIZE ) {
			buffer_.getline(line);
			error_ = rrp_request_error;
			return false;
		} else {
			error_ = rrp_part_command;
			return false;
		}
	} else if( line.size() > MAX_LINE_SIZE ) {
		error_ = rrp_request_error;
		return false;
	} else {
		string command;
		string parameter;
		string protocol_version;
		protocol_version = SplitWord(line);
		command = SplitWord(line);
		parameter = string( find_if(line.begin(), line.end(), NotSpace), line.end() );
		if( protocol_version.empty() || command.empty() || parameter.empty() ) {
			error_ = rrp_request_error;
			return false;
		}
		if( !ParseProtocolVersion(protocol_version, pRRData) ) {
			error_ = rrp_request_error;
			return false;
		}
		pRRData->command.name = command;
		pRRData->command.parameter = parameter;
		error_ = rrp_no_error;
		return true; 
	}
}

// ========================================================================================
///////////////////////////////////////////////////////////////////////////////////////////
